/*

 VGAFONT.C - Turbo C++ 3.0 DOS program
 ---------

 A simple program to display the text-mode (alphanumeric display)
 VGA-BIOS fonts, and dump to a hexfile 'vgafontc.out'

 The program only works with fonts whose width = 8 pixels (8x8, 8x14, 8x16)
 It does not work with font-widths of 9-pixels (9x14, 9x16.)

 The output file is a raw binary dump of the requested font-table, copied
 from the VGABIOS ROM.  The format is approximately as follows :

 typedef struct VGA_CHAR_STRUCT {
        unsigned char pixelrow[ CHAR_HEIGHT ];
        // BITMAP FORMAT :
        // least-significant bit of pixelrow = right-most pixel,
        // most-significant bit of pixelrow = left-most pixel;
        // Rows are stored in top -> bottom order (top=row0 is stored first.)
 } VGA_CHAR;

 typedef struct VGAFONT_STRUCT {
        VGA_CHAR characters[ 256 ];
        // Characters are stored in ASCII order (0,1,2,3,4, ... 255)
 } VGAFONT;

 // CHAR_HEIGHT = 8, 14, or 16

 // Total size of memory dump (in bytes) = CHAR_HEIGHT * 256

 This program comes with no warranty.

 --------------------------

 USAGE :

       If running VGAFONTC in a full-screen DOS-box, run this command first :

          MODE co80,50

       This will set the text-mode to an 80x50 display, enabling VGAFONTC
       to fit more characters on the screen.  (And the resulting aspect
       ratio is closer to actual proportion.)

       VGAFONTC.EXE 2 14 - copy EGA(8x14) fonttable to 'vgafontc.out'

       VGAFONTC.EXE 3 8  - copy CGA(8x8) font-table to 'vgafontc.out'

       VGAFONTC.EXE 6 16 - copy VGA(8x16) font-table to 'vgafonc.out'

 --------------------------

 NOTES :

   Borland's free C++ compiler should be able to compile this program.

 Useful links :
   http://home.worldonline.dk/~finth - VGADOC4B homepage
   http://www.borland.com - free Borland C++ compiler

  liaor@iname.com http://members.tripod.com/~liaor  11/2000

 */




#include<dos.h>
#include<stdio.h>
#include<stdlib.h>

#define VIDEO_INTERRUPT_ID 0x10 /* MS-DOS systems, int86 0x10 = VGA BIOS */

#define FONT_CUR1 0x00 /* current FONT in vector 0x1F */
#define FONT_CUR2 0x01 /* current FONT in vector 0x43 */
#define FONT_8X14 0x02 /* EGA 640x350 -> 80x25 text */
#define FONT_8X8  0x03 /* CGA 320x200 -> 40x25 text */
#define FONT_9X14 0x05 /*     720x350 -> 80x25 text */
#define FONT_8X16 0x06 /* VGA 640x480 -> 80x30 text */
#define FONT_9X16 0x07 /* VGA 720x400 -> 80x25 text */

/*extern int _argc;    */
/*extern char *_argv[];*/

/* print an 8-pixel wide character */
void
print_char8( char far *pfont, const int char_height )
{
     int j, bit;
     unsigned char bitreg = 0;

     /* bitreg = [76543210], bit7 = 'left-most pixel' */

     printf("\n\t    01234567");
     for ( j = 0; j < char_height; j = j + 1)
     {
          printf( "\n\t%02d |", j);
          bitreg = pfont[j];
          for ( bit=0; bit < 8; bit=bit+1 )
          {
               if ( bitreg & 0x80 )
                  printf( "%c", 254 ); /* ANSI 254 = '' */
               else
                   printf( " " );
               bitreg = bitreg << 1;
          }
          printf( "|");
     }

}

void
print_font( char far *pfont )
{
     int i, j, bit;
     const int count = 5; /* number of characters to print */
     const int char_height = 16;
     char far *pchar;

     for ( i = 0; i < count; i = i + 1 )
     {
          printf("\n\nCHARACTER #%d", i );
          pchar = &pfont[ i*char_height ];

          print_char8( pchar, char_height );
     }
}


int
write_hexfile( const char buffer[], const char filename[], const int count )
{
     FILE *fpout = NULL;
     int n = 0; /* count of bytes written to disk */

     fpout = fopen( filename, "rb" );
     if ( fpout != NULL )
     {
          printf("\nwrite_hexfile() : WARNING '%s' already exists, overwriting!",
               filename );
          fclose( fpout );
     }

     fpout = fopen( filename, "wb+" ); /* write binary, overwrite existing */
     if ( fpout == NULL )
     {
          printf("\nwrite_hexfile() : ERROR Could not write to '%s'!",
               filename );
          return -1;
     }

     n = fwrite( (const void *) buffer, sizeof(char), count, fpout );
     if ( n != count )
     {
          printf("\nwrite_hexfile() : WARNING fwrite returned %d instead of %d!",
               n, count );
     }

     fclose( fpout );

     return 0;
}


/* print font-file in XESS '.HEX' format */
int
write_txtfile( const char buffer[], const char filename[], const int count )
{
     FILE *fpout = NULL;
     unsigned int n = 0; /* count of bytes written to disk */
     unsigned int i; /* loop variable counter */
     unsigned int bpl=16; /* bytes per line */

     fpout = fopen( filename, "rt" );
     if ( fpout != NULL )
     {
          printf("\nwrite_txtfile() : WARNING '%s' already exists, overwriting!",
               filename );
          fclose( fpout );
     }

     fpout = fopen( filename, "wt+" ); /* write text, overwrite existing */
     if ( fpout == NULL )
     {
          printf("\nwrite_hexfile() : ERROR Could not write to '%s'!",
               filename );
          return -1;
     }

     /* must typecast variables to 'unsigned', otherwise printf %0X will
        print the 2's complement equivalent of negative values. */
     for ( n = 0; n < count; n = n + bpl )
     {
          fprintf( fpout, "- %02X %04X ", bpl, (unsigned) n );
          for ( i = 0; i < bpl; i = i + 1)
              fprintf( fpout, "%02X ", (unsigned char) buffer[ n + i] );
          fprintf( fpout, "\n" );
     }

     fclose( fpout );

     return 0;
}




int
write_font( char far *pfont, const char filename[], const char filename2[],
            const int count )
{
     int i;
     char *pmem = NULL;

     if ( count <= 0 )
     {
        printf("\nwrite_font( '%s', %d ) : ERROR count < 0 !", filename,
             count );
        return -1;
     }

     /* 1) allocate local memory */
     pmem = (char *)malloc( count*sizeof(char) );
     if ( pmem == NULL )
     {
        printf("\nwrite_font( '%s', %d ) : ERROR malloc() failed!", filename,
             count );
        return -1;
     }

     /* 2) copy memory */
     for ( i = 0; i < count; i=i+1 )
     {
          pmem[ i ] = pfont[ i ];
     }

     if ( filename2 != NULL )
        i = write_txtfile( pmem, filename2, count );

     if ( filename != NULL )
        i = write_hexfile( pmem, filename, count );

     /* 4) deallocate memory */
     free( pmem );

     return i;
}
int
main( void )
{
  struct REGPACK rp;  /* 8086 register state */
  short font_offset, font_segment;
  int i, r;
  int char_height = 16; /* default character height, 16 lines */
  int font_id = 0x0600; /* default FontID, 8x16 */

  char far *pfont = NULL ; /* pointer to vga font*/
  char *font_filename = "vgafontc.out";
  char *font_filename_txt = "vgafontc.hex";

  /* parse command line arguments */

  if ( _argc > 1 ) /* get font ID# */
  {
     i = atoi( _argv[ 1 ] );

     if ( i < 0 || i > 255 )
        printf("\n'%d' : ILLEGAL VALUE for FONT#, using default.", i );
     else
        font_id = i << 8;
  }

  if ( _argc > 2 )
  {
     i = atoi( _argv[ 2 ] );

     if ( i <= 0 )
        printf("\n'%d' : ILLEGAL VALUE for char_height, using default.", i );
     else
        char_height = i;
  }

  if ( _argc > 3 )
  {
     font_filename = _argv[ 3 ] ; /* set pointer to _argv[3] */
  }

/*  rp.r_bx = 0x0600;    BH = 0x06, BL=0xXX, 8x16 font table */

  /* subfunction 30H, get poininter information to font */
  /* Please refer to VGADOC4B for more information */

  rp.r_ax = 0x1130;
  rp.r_bx = font_id;

  printf("\nCalling intr( 0x10, BX=0x%4X ... );", rp.r_bx );

  intr( VIDEO_INTERRUPT_ID, &rp ); /* Activate software interrupt */
  font_offset  = rp.r_bp;
  font_segment = rp.r_es;


  /* 2) create a far pointer from the returned ES:BP address */
  pfont = MK_FP( font_segment, font_offset );
  printf("\nGot font_segment:offset = 0x%0X:%0X\n", font_segment,
                font_offset );

  printf("\nFONT DIMENSIONS : 8 X %d\n-----------", char_height );

  /* 3) write 256 chars of the font to disk */
  write_font( pfont, font_filename, font_filename_txt, char_height*256 );

  i = 0;
  do {
     printf("\nChar ##(decimal {0..255}, '-1' exits) : " );
     r = scanf( " %d", &i ); /* if r != 1, scanf() failed */
     if ( i >= 0 && r == 1)
        print_char8( &pfont[ i * char_height ], char_height );
  } while ( i >= 0 && r == 1);

  //print_font( pfont );

  return 0;
}